import axinite as ax
import numpy as np
import astropy.units as u
from astropy.coordinates import CartesianRepresentation
from numba import jit

def load(delta, limit, backend, *bodies, t=0 * u.s, modifier=None, action=None, action_frequency=200):
    """Loads a simulation from a backend.

    Args:
        delta (u.Quantity): The change in time between each step.
        limit (u.Quantity): The limit of the simulation.
        backend (function): The backend to load the simulation from.
        t (u.Quantity, optional): The initial time. Defaults to 0 * u.s.
        modifier (function, optional): The modifier to apply to forces in the simulation. Defaults to None.
        action (function, optional): An action to call with frequency `action_frequency`. Defaults to None.
        action_frequency (int, optional): The frequency at which to call the `action`. Defaults to 200.

    Returns:
        np.ndarray: An array of bodies.
    """
    body_dtype = np.dtype([
        ("m", np.float64),
        ("r", np.float64, (int(limit.value/delta.value), 3)),
        ("v", np.float64, (int(limit.value/delta.value), 3))
    ])
    _bodies = np.array([], dtype=body_dtype)
    for body in bodies:
        _r = body.r.values()
        _v = body.v.values()
        r = np.zeros((int(limit.value/delta.value), 3))
        v = np.zeros((int(limit.value/delta.value), 3))
        for i, __r in enumerate(_r): 
            r[i][0] = __r.x.value
            r[i][1] = __r.y.value
            r[i][2] = __r.z.value
        for i, __v in enumerate(_v):
            v[i][0] = __v.x.value
            v[i][1] = __v.y.value
            v[i][2] = __v.z.value
        _bodies = np.append(_bodies, np.array([
            (body.mass.value, r, v)
        ], dtype=body_dtype))
    _bodies = backend(delta.value, limit.value, _bodies, action=action, modifier=modifier, t=t.value, action_frequency=action_frequency)
    __bodies = ()
    for body in _bodies: 
        _body = ax.Body(body["m"] * u.kg, CartesianRepresentation(*body["r"][0], u.m), CartesianRepresentation(*body["v"][0], u.m/u.s))
        for i, r in enumerate(body["r"]):
            _body.r[i * delta.value] = CartesianRepresentation(*r, u.m)
        for i, v in enumerate(body["v"]):
            _body.v[i * delta.value] = CartesianRepresentation(*v, u.m/u.s)
        __bodies += (_body,)
    return __bodies



from astropy.coordinates import CartesianRepresentation
from astropy.constants import G
import axinite as ax
import astropy.units as u
from numpy import float64

class Body:
    "A class that represents a body in the simulation."
    def __init__(self, name: str, mass: u.Quantity, position: CartesianRepresentation, velocity: CartesianRepresentation, radius: u.Quantity, color: str = "", light: bool = False, retain = None, radius_multiplier = 1):
        """Initializes a new Body object.

        Args:
            name (str): The body's name.
            mass (u.Quantity): The mass of the body in kilograms.
            position (CartesianRepresentation): The initial position of the body (in vector form).
            velocity (CartesianRepresentation): The initial velocity of the body (in vector form).
            radius (u.Quantity): The radius of the body in meters.
            color (str, optional): The color of the body. Defaults to "", telling the frontend to choose from a list.
            light (bool, optional): Whether the body should give of light. Defaults to False.
            retain (_type_, optional): How many points the body should retain on it's trail. Defaults to None, meaning "go with default".
            radius_multiplier (int, optional): A multiplier to be applied to the radius. Defaults to 1.
        """

        self.mass: u.Quantity = mass
        "The mass of the body in kilograms."

        self.r: dict[float64, CartesianRepresentation] = { float64(0): position}
        "The position of the body at each timestep."

        self.v: dict[float64, CartesianRepresentation] = { float64(0): velocity}
        "The velocity of the body at each timestep."

        self.name: str = name
        "The name of the body."

        self.radius: u.Quantity = radius * radius_multiplier
        "The radius of the body in meters."

        self.color: str = color
        "The color of the body."

        self.light: bool = light
        "Whether the body should give off light."

        self.retain: int = retain
        "How many points the body should retain on it's trail."

        self.radius_multiplier: int = radius_multiplier
        "A multiplier which was applied to the radius."


from astropy.coordinates import CartesianRepresentation
from astropy.units import Quantity
from astropy.constants import G
import astropy.units as u
import math
import numpy as np
from numba import njit

_G = G.value

def apply_to_vector(vector: CartesianRepresentation, function) -> CartesianRepresentation:
    """Applies a function to each component of a vector.

    Args:
        vector (CartesianRepresentation): The vector to apply the function to.
        function (_type_): The function to apply to the vector.

    Returns:
        CartesianRepresentation: The vector with the function applied to it.
    """
    return CartesianRepresentation([function(i) for i in vector.xyz])

def vector_to(vector: CartesianRepresentation, unit: Quantity) -> CartesianRepresentation:
    """Converts a vector to a specific unit.

    Args:
        vector (CartesianRepresentation): The vector to convert.
        unit (Quantity): The unit to convert the vector to.

    Returns:
        CartesianRepresentation: The vector in the new unit.
    """
    return apply_to_vector(vector, lambda i: i.to(unit))

def vector_magnitude(vector: CartesianRepresentation) -> float:
    """Calculates the magnitude of a vector.

    Args:
        vector (CartesianRepresentation): The vector to calculate the magnitude of.

    Returns:
        float: The magnitude of the vector.
    """
    return np.sqrt(vector.x**2 + vector.y**2 + vector.z**2)

def unit_vector(vector: CartesianRepresentation) -> CartesianRepresentation:
    """Calculates the unit vector of a vector.

    Args:
        vector (CartesianRepresentation): The vector to calculate the unit vector of.

    Returns:
        CartesianRepresentation: The unit vector of the vector.
    """
    return vector / vector_magnitude(vector)

def to_vector(data: dict, unit: u.Unit) -> CartesianRepresentation:
    """Converts a dictionary to a vector.

    Args:
        data (dict): The dictionary to convert.
        unit (u.Unit): The unit to convert the vector to.

    Returns:
        CartesianRepresentation: The vector.
    """
    return CartesianRepresentation(data["x"] * unit, data["y"] * unit, data["z"] * unit)

@njit 
def vector_magnitude_jit(vec: np.ndarray) -> float:
    """Calculates the magnitude of a vector.

    Args:
        vec (np.ndarray): The vector to calculate the magnitude of.

    Returns:
        float: The magnitude of the vector.
    """
    return np.sqrt(np.sum(vec**2))

@njit
def unit_vector_jit(vec: np.ndarray) -> np.ndarray:
    """Calculates the unit vector of a vector.

    Args:
        vec (np.ndarray): The vector to calculate the unit vector of.

    Returns:
        np.ndarray: The unit vector of the vector.
    """
    mag = vector_magnitude_jit(vec)
    return vec / mag if mag > 0 else vec

@njit
def gravitational_force_jit(m1: u.Quantity, m2: u.Quantity, r: np.ndarray) -> np.ndarray:
    """Calculates the gravitational force between two bodies.

    Args:
        m1 (u.Quantity): The mass of the first body.
        m2 (u.Quantity): The mass of the second body.
        r (np.ndarray): The vector between the two bodies.

    Returns:
        np.ndarray: The gravitational force between the two bodies.
    """
    mag = vector_magnitude_jit(r)
    if mag == 0:
        return np.zeros(3)
    return -_G *((m1 * m2) / mag**2) * unit_vector_jit(r)


import axinite as ax
from vpython import *
import axinite.tools as axtools
from itertools import cycle
import vpython as vp

colors = cycle([color.red, color.blue, color.green, color.orange, color.purple, color.yellow])

def show(_args: axtools.AxiniteArgs, frontend: 'function') -> None:
    """Statically display the bodies in the simulation.

    Args:
        _args (axtools.AxiniteArgs): The simulation parameters.
        frontend (function): The frontend function.
    """

    args = _args
    if args.rate is None:
        args.rate = 100
    if args.radius_multiplier is None:
        args.radius_multiplier = 1
    if args.retain is None:
        args.retain = 200

    for body in args.bodies:
        frontend[0](body)

    if frontend[1] != None: 
        frontend[1]()

    try:
        while True:
            pass
    except KeyboardInterrupt:
        frontend[2]()


import axinite.tools as axtools
import json

def save(args: axtools.AxiniteArgs, path: str):
    with open(path, 'w+') as f:
        data = {
            "name": args.name,
            "delta": args.delta.value,
            "limit": args.limit.value,
            "t": args.t.value,
            "radius_multiplier": args.radius_multiplier,
            "bodies": []
        }

        for body in args.bodies: 
            body_data = {
                "name": body.name,
                "mass": body.mass.value,
                "radius": body.radius.value,
                "r": {k: [v.x.value, v.y.value, v.z.value] for k, v in body.r.items()},
                "v": {k: [v.x.value, v.y.value, v.z.value] for k, v in body.v.items()}
            }
            if body.color != None:
                body_data["color"] = body.color
            if body.retain != None:
                body_data["retain"] = body.retain
            if body.light != None:
                body_data["light"] = body.light

            data["bodies"].append(body_data)

        if args.radius_multiplier is not None:
            data["radius_multiplier"] = args.radius_multiplier

        if args.rate is not None:
            data["rate"] = args.rate

        if args.retain is not None:
            data["retain"] = args.retain
        
        if args.frontend_args != {}:
            data["frontend_args"] = args.frontend_args

        json.dump(data, f, indent=4)


import axinite.tools as axtools
import plotly.graph_objects as go
from itertools import cycle
import os, signal

def plotly_frontend(args: axtools.AxiniteArgs, mode: str, theme="plotly_dark", use_min=True):
    if mode != "show": raise Exception("plotly_frontend is only supported in show mode.")
    if args.rate is None:
        args.rate = 100
    if args.radius_multiplier is None:
        args.radius_multiplier = 1
    if args.retain is None:
       args.retain = 200
    if args.frontend_args != {} and "plotly" in args.frontend_args:
        if "theme" in args.frontend_args["plotly"]:
            theme = args.frontend_args["plotly"]["theme"]
        if "use_min" in args.frontend_args["plotly"]:
            use_min = args.frontend_args["plotly"]["use_min"]

    minlen = axtools.min_axis_length(*args.bodies, radius_multiplier=args.radius_multiplier)
    maxlen = axtools.max_axis_length(*args.bodies, radius_multiplier=args.radius_multiplier)

    range = [minlen if use_min else -maxlen, maxlen]

    colors = cycle(['red', 'blue', 'green', 'orange', 'purple', 'yellow'])

    fig = go.Figure()
    layout = go.Layout(
        autosize=True,
        template=theme,
        scene=dict(
            aspectmode='cube',
            xaxis=dict(range=range),
            yaxis=dict(range=range),
            zaxis=dict(range=range)
        )
    )
    fig.update_layout(layout)

    def fn(body: axtools.Body):
        body_color = body.color if body.color != "" else next(colors)
        xx, yy, zz = axtools.create_sphere(body.r[0], body.radius * body.radius_multiplier)
        fig.add_trace(go.Surface(
            x=xx,
            y=yy,
            z=zz,
            colorscale=[[0, body_color], [1, body_color]],
            cauto=False,
            cmin=0,
            cmax=1,
            showlegend=True,
            showscale=False,
            name=body.name,
            opacity=0.2
        ))
        fig.add_trace(go.Scatter3d(
            x=[body.r[0].x.value],
            y=[body.r[0].y.value],
            z=[body.r[0].z.value],
            mode='text',
            text=[body.name],
            textposition='top center',
            showlegend=False
        ))
        trajectory_x = [point.x.value for point in body.r.values()]
        trajectory_y = [point.y.value for point in body.r.values()]
        trajectory_z = [point.z.value for point in body.r.values()]
        fig.add_trace(go.Scatter3d(
            x=trajectory_x,
            y=trajectory_y,
            z=trajectory_z,
            mode='lines',
            line=dict(color=body_color, width=2),
            name=f"{body.name} trajectory",
            showlegend=False
        ))

        
    return fn, fig.show, lambda: os.kill(os.getpid(), signal.SIGINT)


import axinite.tools as axtools
from vpython import *
from itertools import cycle
from astropy.coordinates import CartesianRepresentation
import os, signal
  
def to_vec(cartesian_representation: CartesianRepresentation):
    return vector(cartesian_representation.x.value, cartesian_representation.y.value, cartesian_representation.z.value)
    
def vpython_frontend(args: axtools.AxiniteArgs, mode: str, **kwargs):
    if mode == "live" or mode == "run":
        return vpython_rt(args, **kwargs)
    elif mode == "show":
        return vpython_static(args, **kwargs)
    
def vpython_rt(args: axtools.AxiniteArgs):
    if args.rate is None:
        args.rate = 100
    if args.radius_multiplier is None:
        args.radius_multiplier = 1
    if args.retain is None:
       args.retain = 200
    
    scene = canvas(title=args.name)
    scene.select()

    colors = cycle([color.red, color.blue, color.green, color.orange, color.purple, color.yellow])

    global pause
    pause = False
    def pause_handler():
        global pause
        pause = not pause

    global _rate
    _rate = args.rate
    def rate_hander(evt):
        global _rate
        _rate = evt.value
    button(text='Pause', bind=pause_handler, pos=scene.caption_anchor)
    slider(bind=rate_hander, min=1, max=1000, value=+_rate, step=1, right=15, length=200, pos=scene.caption_anchor)

    spheres = {}
    labels = {}
    lights = {}

    for body in args.bodies:
        body_color = axtools.string_to_color(body.color, "vpython") if body.color != "" else next(colors)
        body_retain = body.retain if body.retain != None else args.retain
        spheres[body.name] = sphere(pos=to_vec(body.r[0]), radius=body.radius.value * args.radius_multiplier, color=body_color, make_trail=True, retain=body_retain, interval=10)
        labels[body.name] = label(pos=spheres[body.name].pos, text=body.name, xoffset=15, yoffset=15, space=30, height=10, border=4, font='sans')
        if body.light == True: 
            lights[body.name] = local_light(pos=to_vec(body.r[0]), color=body_color)
            attach_light(spheres[body.name], lights[body.name])

    def fn(t, **kwargs):
        try:
            global _rate, pause
            rate(_rate)
            for body in kwargs["bodies"]:
                spheres[body.name].pos = to_vec(body.r[t.value])
                labels[body.name].pos = spheres[body.name].pos
                try: lights[body.name].pos = spheres[body.name].pos
                except: pass
            print(f"t = {t}", end='\r')
            if pause: 
                while pause: rate(10)
        except KeyboardInterrupt:
            os.kill(os.getpid(), signal.SIGINT)

    return fn, lambda: os.kill(os.getpid(), signal.SIGINT)

def vpython_static(args: axtools.AxiniteArgs):
    if args.rate is None:
        args.rate = 100
    if args.radius_multiplier is None:
        args.radius_multiplier = 1
    if args.retain is None:
       args.retain = 200
       
    scene = canvas(title=args.name)
    scene.select()

    colors = cycle([color.red, color.blue, color.green, color.orange, color.purple, color.yellow])

    def fn(body: axtools.Body):
        body_color = axtools.string_to_color(body.color, "vpython") if body.color != "" else next(colors)
        
        label(pos=to_vec(body.r[0]), text=body.name, xoffset=15, yoffset=15, space=30, height=10, border=4, font='sans', color=body_color)
        curve(pos=[to_vec(r) for r in body.r.values()], color=body_color)
        sphere(pos=to_vec(body.r[0]), radius=body.radius.value * args.radius_multiplier, color=body_color, opacity=0.2, make_trail=False)

    return fn, None, lambda: os.kill(os.getpid(), signal.SIGINT)


import axinite as ax
import axinite.tools as axtools
from axinite.tools import AxiniteArgs
import json
from numba import jit

def load(args: AxiniteArgs, path: str = "", dont_change_args: bool = False, verbose: bool = True):
    """Preloads a simulation.

    Args:
        args (AxiniteArgs): An AxiniteArgs object containing simulation parameters.
        path (str, optional): The path to dump the computed simulation to. Defaults to "", which skips data dumping.
        dont_change_args (bool, optional): True if the load function should take care not to edit the args object. Defaults to False.
        jit (bool, optional): Whether JIT should be used. Defaults to True.
        verbose (bool, optional): Whether the load function should print to console. Defaults to True.

    Returns:
        list[axtools.Body]: A list of Body objects containing the preloaded simulation data.
    """
    
    if args.action == None: 
        delta = args.delta.value
        limit = args.limit.value

        @jit(nopython=False)
        def default_action(bodies, t): print("Timestep", round(t / delta), "(", t / limit * 100, "\b% )")

        args.action = default_action if verbose else None
        args.action_frequency = 200
    if args.backend == None: args.backend = ax.verlet_backend

    bodies = ax.load(*args.unpack(), t=args.t, modifier=args.modifier, action=args.action)
    if verbose: print(f"Finished with {len(bodies[0].r)} timesteps")

    _bodies = []
    for i, body in enumerate(bodies):
        _bodies.append(args.bodies[i])
        for j, r in body.r.items():
            _bodies[i].r[j] = r
            _bodies[i].v[j] = body.v[j]
    
    if not dont_change_args:
        args.t = args.limit
        args.bodies = _bodies
    if path == "":
        return _bodies
    else: 
        axtools.save(args, path)
        return _bodies


import axinite as ax
import axinite.tools as axtools
from vpython import *
from itertools import cycle
import astropy.units as u

colors = cycle([color.red, color.blue, color.green, color.orange, color.purple, color.yellow])

def live(args: axtools.AxiniteArgs, frontend: 'function') -> None:
    """Watch a preloaded simulation live.

    Args:
        args (axtools.AxiniteArgs): The arguments for the simulation.
        frontend (function): The frontend to use.
    """
    if args.rate is None:
        args.rate = 100
    if args.radius_multiplier is None:
        args.radius_multiplier = 1
    if args.retain is None:
       args.retain = 200

    t = 0 * u.s
    try:
        while t < args.limit:
            frontend[0](t, bodies=args.bodies)
            t += args.delta
    finally: frontend[1]()


import astropy.units as u
import axinite as ax
from astropy.coordinates import CartesianRepresentation
from typing import Literal
import vpython as vp
import numpy as np
import axinite.tools as axtools

def interpret_time(string: str) -> u.Quantity:
    """Interprets a string as a time in seconds.

    Args:
        string (str): The string to interpret.

    Returns:
        u.Quantity: The time in seconds.
    """
    if type(string) is float: return string * u.s
    if string.endswith("min"):
        string = string.removesuffix("min")
        return float(string) * 60 * u.s 
    elif string.endswith("hr"): 
        string = string.removesuffix("hr")
        return float(string) * 3600 * u.s
    elif string.endswith("d"):
        string  = string.removesuffix("d")
        return float(string) * 86400 * u.s
    elif string.endswith("yr"):
        string = string.removesuffix("yr")
        return float(string) * 31536000 * u.s
    else: return float(string) * u.s

def array_to_vectors(array: list[dict[str, np.float64]], unit: u.Unit) -> list[CartesianRepresentation]:
    """Converts a list of dicts to a list of vectors.

    Args:
        array (list[dict[str, np.float64]]): A list of dicts to convert.
        unit (u.Unit): The unit to convert to.

    Returns:
        list[CartesianRepresentation]: The list of vectors.
    """
    arr = []
    for a in array:
        arr.append(ax.to_vector(a, unit))
    return arr

def data_to_body(data: dict[str, any]) -> axtools.Body:
    """Converts a dict to a Body object.

    Args:
        data (dict[str, any]): The dict to convert.

    Returns:
        axtools.Body: The Body object.
    """

    name = data["name"]
    mass = data["mass"] * u.kg
    
    if "x" in data["r"]:
        position = ax.to_vector(data["r"], u.m)
        velocity = ax.to_vector(data["v"], u.m/u.s)

        body = axtools.Body(name, mass, position, velocity, data["radius"] * u.m)

        if "color" in data:
            body.color = data["color"]
        if "light" in data:
            body.light = data["light"]
        if "retain" in data:
            body.retain = data["retain"]

        return body
    else:
        position = [vector_from_list(r, u.m) for r in data["r"].values()]
        velocity = [vector_from_list(v, u.m/u.s) for v in data["v"].values()]

        body = axtools.Body(name, mass, position[0], velocity[0], data["radius"] * u.m)

        for t, r in data["r"].items():
            body.r[to_float(t)] = vector_from_list(r, u.m)
        for t, v in data["v"].items():
            body.v[to_float(t)] = vector_from_list(v, u.m)

        if "color" in data:
            body.color = data["color"]
        if "light" in data:
            body.light = data["light"]
        if "retain" in data:
            body.retain = data["retain"]
        if "radius_multiplier" in data:
            body.radius *= data["radius_multiplier"]
        
        return body

def vector_from_list(vector: list[np.float64], unit: u.Unit) -> CartesianRepresentation:
    """Converts a list to a vector.

    Args:
        vector (list[np.float64]): The list to convert
        unit (u.Unit): The unit to convert to.

    Returns:
        CartesianRepresentation: The vector.
    """

    return CartesianRepresentation(u.Quantity(float(vector[0]), unit), u.Quantity(float(vector[1]), unit), u.Quantity(float(vector[2]), unit))

def to_float(val: any) -> np.float64:
    """Legacy shorthand for converting to np.float64.

    Args:
        val (any): The value to convert.

    Returns:
        np.float64: The result.
    """

    return np.float64(val)

def string_to_color(color_name: str, frontend: Literal['vpython', 'mpl', 'plotly']) -> vp.color | str:
    """Converts a string to a color object for a given frontend.

    Args:
        color_name (str): The name of the color.
        frontend (str): The frontend to convert for.

    Returns:
        vp.color | str: The converted color.
    """
    if frontend == "vpython":
        color_map = {
            'red': vp.color.red,
            'blue': vp.color.blue,
            'green': vp.color.green,
            'orange': vp.color.orange,
            'purple': vp.color.purple,
            'yellow': vp.color.yellow,
            'white': vp.color.white,
            'gray': vp.color.gray(0.5)
        }
        return color_map.get(color_name, vp.color.white)
    elif frontend == "matplotlib":
        color_map = {
            'red': 'r',
            'blue': 'b',
            'green': 'g',
            'orange': 'orange',
            'purple': 'purple',
            'yellow': 'yellow',
            'white': 'white',
            'gray': 'gray'
        }
        return color_map.get(color_name, 'white')
    

def create_sphere(pos: CartesianRepresentation, radius: u.Quantity, n=20) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Generates the vertices of a sphere.

    Args:
        pos (CartesianRepresentation): The position of the sphere.
        radius (u.Quantity): The radius of the sphere.
        n (int, optional): Number of segments used to generate verticies. Defaults to 20.

    Returns:
        tuple[np.ndarray, np.ndarray, np.ndarray]: The x, y, and z coordinates of the sphere.
    """
    u1 = np.linspace(0, 2 * np.pi, n)
    v1 = u1.copy()
    uu, vv = np.meshgrid(u1, v1)

    xx = pos.x.value + radius.value * np.cos(uu) * np.sin(vv)
    yy = pos.y.value + radius.value * np.sin(uu) * np.sin(vv)
    zz = pos.z.value + radius.value * np.cos(vv)

    return xx, yy, zz

def max_axis_length(*bodies: axtools.Body, radius_multiplier: int = 1) -> np.float64:
    """Finds the maximum axis length of a set of bodies.

    Args:
        radius_multiplier (int, optional): The radius multiplier to apply. Defaults to 1.

    Returns:
        np.float64: The longest axis length found.
    """

    max_length = 0
    for body in bodies:
        x_length = max([v.x.value for k, v in body.r.items()]) + body.radius.value * radius_multiplier
        y_length = max([v.y.value for k, v in body.r.items()]) + body.radius.value * radius_multiplier
        z_length = max([v.z.value for k, v in body.r.items()]) + body.radius.value * radius_multiplier
        
        max_length = max(max_length, x_length, y_length, z_length)
    
    return max_length

def min_axis_length(*bodies: axtools.Body, radius_multiplier: int = 1) -> np.float64:
    """Finds the minimum axis length of a set of bodies.

    Args:
        radius_multiplier (int, optional): The radius multiplier to apply. Defaults to 1.

    Returns:
        np.float64: The lowest axis length found.
    """
    
    min_length = 0
    for body in bodies:
        x_length = min([v.x.value for k, v in body.r.items()]) - body.radius.value * radius_multiplier
        y_length = min([v.y.value for k, v in body.r.items()]) - body.radius.value * radius_multiplier
        z_length = min([v.z.value for k, v in body.r.items()]) - body.radius.value * radius_multiplier
        
        min_length = min(min_length, x_length, y_length, z_length)
    
    return min_length


from astropy.coordinates import CartesianRepresentation
from astropy.constants import G
from axinite.functions import vector_to, apply_to_vector, vector_magnitude, unit_vector
import astropy.units as u
from math import pi
from numpy import float64

class Body:
    def __init__(self, mass: u.Quantity, position: CartesianRepresentation, velocity: CartesianRepresentation):
        """Initializes a new Body object.

        Args:
            mass (u.Quantity): Mass of the object in kilograms.
            position (CartesianRepresentation): The initial position of the object.
            velocity (CartesianRepresentation): The initial velocity of the object.
        """
        self.mass = mass
        "The mass of the object in kilograms."

        self.r = { float64(0): position}
        "The position of the object at each timestep."

        self.v = { float64(0): velocity}
        "The velocity of the object at each timestep."


from axinite.tools.args import AxiniteArgs
from axinite.tools.body import Body
from axinite.tools.functions import interpret_time, array_to_vectors, data_to_body, vector_from_list, to_float, \
    create_sphere, max_axis_length, min_axis_length
from axinite.tools.load import load
from axinite.tools.read import read
from axinite.tools.show import show
from axinite.tools.live import live
from axinite.tools.run import run
from axinite.tools.frontends.vpython import vpython_frontend
from axinite.tools.frontends.plotly import plotly_frontend
from axinite.tools.save import save


import numpy as np
from numba import njit, typed, types, jit
import axinite as ax

def euler_nojit_backend(delta, limit, bodies, action=None, modifier=None, t=-1.0, action_frequency=200):
    _infinite = False
    if t == 0.0: t = 0.0 + delta
    if t == -1.0: 
        _infinite = True
        t = 0.0 + delta
        
    n = 1
    while t < limit or _infinite:
        for i, body in enumerate(bodies):
            f = np.zeros(3)
            for j, other in enumerate(bodies):
                if i != j:
                    r = body["r"][n - 1] - other["r"][n - 1]
                    f += ax.gravitational_force_jit(body["m"], other["m"], r)
            if modifier is not None: f = modifier(body, f)
            a = f / body["m"]
            v = body["v"][n - 1] + a * delta
            r = body["r"][n - 1] + v * delta
            body["v"][n] = v
            body["r"][n] = r
        t += delta
        n += 1
        if action is not None and n % action_frequency == 0: action(bodies, t)
    return bodies

@jit(nopython=False)
def euler_backend(delta, limit, bodies, action=None, modifier=None, t=-1.0, action_frequency=200):
    return jit(euler_nojit_backend, nopython=True)(delta, limit, bodies, action=action, modifier=modifier, t=t, action_frequency=action_frequency)


import axinite as ax
import axinite.tools as axtools
from vpython import *

def run(_args: axtools.AxiniteArgs, frontend: 'function') -> tuple[axtools.Body, ...]:
    """Load and display a simulation simultaneously.

    Args:
        _args (axtools.AxiniteArgs): The simulation parameters.
        frontend (function): The frontend to use.

    Returns:
        tuple[axtools.Body, ...]: The bodies in the simulation
    """

    args = _args
    if args.rate is None:
        args.rate = 100
    if args.radius_multiplier is None:
        args.radius_multiplier = 1
    if args.retain is None:
       args.retain = 200

    args.action = frontend[0]
    try: bodies = ax.load_legacy(*args.unpack(), t=args.t, action=args.action)
    finally: 
        frontend[1]()
        return bodies


import axinite.tools as axtools
import astropy.units as u

class AxiniteArgs:
    "A class to store simulation parameters."
    def __init__(self):
        """Initializes an AxiniteArgs object."""

        self.name: str = None
        "The name of the simulation."

        self.delta: u.Quantity = None
        "The frequency at which the simulation should be computed in seconds."

        self.limit: u.Quantity = None
        "The length of the simulation in seconds."

        self.action: function = None
        "A function to be called at each timestep."

        self.t: u.Quantity = None
        "The current timestep."

        self.bodies: list[axtools.Body] = []
        "A list of Body objects to be simulated."

        self.radius_multiplier: float = None
        "A float to multiply all radii by."

        self.rate: int = None
        "The number frames per second to render for the live and run functions."

        self.retain: int = None
        "The number of points to retain on each body's trail."

        self.modifier: function = None
        "A function called to modify the forces on the bodies."

        self.frontend_args: dict[str, dict[str, str|float|int|bool|list|dict]] = {}
        "A dictionary of frontend-specific arguments."

        self.backend: function = None
        "The backend (integration method) to use."

        self.action_frequency: int = None
        "The frequency at which the action function should be called."

    def unpack(self) -> tuple[u.Quantity, u.Quantity, 'function', '*tuple[axtools.Body, ...]']:
        """Unpacks the AxiniteArgs object into a tuple that can be passed in `axinite`'s load function.

        Returns:
            *tuple[u.Quantity, u.Quantity, *tuple[axtools.Body, ...]]: An unpackable tuple of the core simulation parameters.
        """
        return self.delta, self.limit, self.backend, *self.bodies


"The `axinite` module is the main module for the Axinite celestial mechanics engine."

from axinite.body import Body
from axinite.functions import apply_to_vector, vector_to, vector_magnitude, unit_vector, to_vector, \
    vector_magnitude_jit, unit_vector_jit, gravitational_force_jit
import axinite.functions as functions
from axinite.load import load
from axinite.backends.euler import euler_backend
from axinite.backends.verlet import verlet_backend